@@ -43,6 +43,7 @@ the local (class-wise) registry for GridGain Complex objects.
 from collections import defaultdict, OrderedDict
 import random
 import re
+from itertools import chain
 from typing import Dict, Iterable, List, Optional, Tuple, Type, Union
 
 from .api.binary import get_binary_type, put_binary_type
@@ -57,7 +58,7 @@ from .exceptions import (
     BinaryTypeError, CacheError, ReconnectError, SQLError, connection_errors,
 )
 from .utils import (
-    capitalize, entity_id, schema_id, process_delimiter, select_version,
+    cache_id, capitalize, entity_id, schema_id, process_delimiter,
     status_to_exception, is_iterable,
 )
 from .binary import GenericObjectMeta
@@ -131,7 +132,14 @@ class Client:
 
     @property
     def partition_aware(self):
-        return self._partition_aware
+        return self._partition_aware and self.partition_awareness_supported_by_protocol
+
+    @property
+    def partition_awareness_supported_by_protocol(self):
+        # TODO: Need to re-factor this. I believe, we need separate class or
+        #  set of functions to work with protocol versions without manually
+        #  comparing versions with just some random tuples
+        return self.protocol_version is not None and self.protocol_version >= (1, 4, 0)
 
     def connect(self, *args):
         """
@@ -158,7 +166,7 @@ class Client:
         # the following code is quite twisted, because the protocol version
         # is initially unknown
 
-        # TODO: open first node in foregroung, others − in background
+        # TODO: open first node in foreground, others − in background
         for i, node in enumerate(nodes):
             host, port = node
             conn = Connection(self, **self._connection_args)
@@ -166,30 +174,18 @@ class Client:
             conn.port = port
 
             try:
-                if (
-                    self.protocol_version is None
-                    or self.protocol_version >= (1, 4, 0)
-                ):
+                if self.protocol_version is None or self.partition_aware:
                     # open connection before adding to the pool
                     conn.connect(host, port)
 
                     # now we have the protocol version
-                    if self.protocol_version < (1, 4, 0):
+                    if not self.partition_aware:
                         # do not try to open more nodes
                         self._current_node = i
-                    else:
-                        # take a chance to schedule the reconnection
-                        # for all the failed connections, that was probed
-                        # before this
-                        for failed_node in self._nodes[:i]:
-                            failed_node.reconnect()
 
             except connection_errors:
-                conn._fail()
-                if (
-                    self.protocol_version
-                    and self.protocol_version >= (1, 4, 0)
-                ):
+                conn.failed = True
+                if self.partition_aware:
                     # schedule the reconnection
                     conn.reconnect()
 
@@ -204,7 +200,6 @@ class Client:
         self._nodes.clear()
 
     @property
-    @select_version
     def random_node(self) -> Connection:
         """
         Returns random usable node.
@@ -213,46 +208,44 @@ class Client:
         extend the `pygridgain` capabilities (with additional testing,
         logging, examining connections, et c.) you probably should not use it.
         """
-        try:
-            return random.choice(
-                list(n for n in self._nodes if n.alive)
-            )
-        except IndexError:
-            # cannot choose from an empty sequence
-            raise ReconnectError('Can not reconnect: out of nodes.') from None
-
-    def random_node_130(self):
-        # it actually returns the next usable node, but the name stands for
-        # the code unification reason
-        node = self._nodes[self._current_node]
-        if node.alive:
-            return node
-
-        # close current (supposedly failed) node
-        self._nodes[self._current_node].close()
-
-        # advance the node index
-        self._current_node += 1
-        if self._current_node >= len(self._nodes):
-            self._current_node = 0
-
-        # prepare the list of node indexes to try to connect to
-        seq = list(range(len(self._nodes)))
-        seq = seq[self._current_node:] + seq[:self._current_node]
-
-        for i in seq:
-            node = self._nodes[i]
+        if self.partition_aware:
+            # if partition awareness is used just pick a random connected node
             try:
-                node.connect(node.host, node.port)
-            except connection_errors:
-                pass
-            else:
+                return random.choice(
+                    list(n for n in self._nodes if n.alive)
+                )
+            except IndexError:
+                # cannot choose from an empty sequence
+                raise ReconnectError('Can not reconnect: out of nodes.') from None
+        else:
+            # if partition awareness is not used then just return the current
+            # node if it's alive or the next usable node if connection with the
+            # current is broken
+            node = self._nodes[self._current_node]
+            if node.alive:
                 return node
 
-        # no nodes left
-        raise ReconnectError('Can not reconnect: out of nodes.')
+            # close current (supposedly failed) node
+            self._nodes[self._current_node].close()
 
-    random_node_120 = random_node_130
+            # advance the node index
+            self._current_node += 1
+            if self._current_node >= len(self._nodes):
+                self._current_node = 0
+
+            # prepare the list of node indexes to try to connect to
+            num_nodes = len(self._nodes)
+            for i in chain(range(self._current_node, num_nodes), range(self._current_node)):
+                node = self._nodes[i]
+                try:
+                    node.connect(node.host, node.port)
+                except connection_errors:
+                    pass
+                else:
+                    return node
+
+            # no nodes left
+            raise ReconnectError('Can not reconnect: out of nodes.')
 
     @status_to_exception(BinaryTypeError)
     def get_binary_type(self, binary_type: Union[str, int]) -> dict:
@@ -521,20 +514,21 @@ class Client:
         return cache_get_names(self.random_node)
 
     def sql(
-        self, query_str: str, page_size: int = 1, query_args: Iterable = None,
-        schema: Union[int, str] = 'PUBLIC',
+        self, query_str: str, page_size: int = 1024,
+        query_args: Iterable = None, schema: str = 'PUBLIC',
         statement_type: int = 0, distributed_joins: bool = False,
         local: bool = False, replicated_only: bool = False,
         enforce_join_order: bool = False, collocated: bool = False,
         lazy: bool = False, include_field_names: bool = False,
         max_rows: int = -1, timeout: int = 0,
+        cache: Union[int, str, Cache] = None
     ):
         """
         Runs an SQL query and returns its result.
 
         :param query_str: SQL query string,
-        :param page_size: (optional) cursor page size. Default is 1, which
-         means that client makes one server call per row,
+        :param page_size: (optional) cursor page size. Default is 1024, which
+         means that client makes one server call per 1024 rows,
         :param query_args: (optional) query arguments. List of values or
          (value, type hint) tuples,
         :param schema: (optional) schema for the query. Defaults to `PUBLIC`,
@@ -561,6 +555,8 @@ class Client:
          (all rows),
         :param timeout: (optional) non-negative timeout value in ms.
          Zero disables timeout (default),
+        :param cache (optional) Name or ID of the cache to use to infer schema.
+         If set, 'schema' argument is ignored,
         :return: generator with result rows as a lists. If
          `include_field_names` was set, the first row will hold field names.
         """
@@ -588,10 +584,13 @@ class Client:
 
         conn = self.random_node
 
-        schema = self.get_or_create_cache(schema)
+        c_id = cache.cache_id if isinstance(cache, Cache) else cache_id(cache)
+
+        if c_id != 0:
+            schema = None
+
         result = sql_fields(
-            conn, schema.cache_id, query_str,
-            page_size, query_args, schema.name,
+            conn, c_id, query_str, page_size, query_args, schema,
             statement_type, distributed_joins, local, replicated_only,
             enforce_join_order, collocated, lazy, include_field_names,
             max_rows, timeout,
