@@ -15,7 +15,7 @@
 #
 from collections import OrderedDict
 import ctypes
-import inspect
+from io import SEEK_CUR
 from typing import Iterable, Dict
 
 from pygridgain.constants import *
@@ -25,15 +25,17 @@ from .internal import AnyDataObject, infer_from_python
 from .type_codes import *
 from .type_ids import *
 from .type_names import *
-
+from .null_object import Null, Nullable
 
 __all__ = [
     'Map', 'ObjectArrayObject', 'CollectionObject', 'MapObject',
     'WrappedDataObject', 'BinaryObject',
 ]
 
+from ..stream import BinaryStream
 
-class ObjectArrayObject(GridGainDataType):
+
+class ObjectArrayObject(GridGainDataType, Nullable):
     """
     Array of GridGain objects of any consistent type. Its Python representation
     is tuple(type_id, iterable of any type). The only type ID that makes sense
@@ -67,15 +69,14 @@ class ObjectArrayObject(GridGainDataType):
         )
 
     @classmethod
-    def parse(cls, client: 'Client'):
+    def parse_not_null(cls, stream):
         header_class = cls.build_header()
-        buffer = client.recv(ctypes.sizeof(header_class))
-        header = header_class.from_buffer_copy(buffer)
-        fields = []
+        header = stream.read_ctype(header_class)
+        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
 
+        fields = []
         for i in range(header.length):
-            c_type, buffer_fragment = AnyDataObject.parse(client)
-            buffer += buffer_fragment
+            c_type = AnyDataObject.parse(stream)
             fields.append(('element_{}'.format(i), c_type))
 
         final_class = type(
@@ -86,10 +87,11 @@ class ObjectArrayObject(GridGainDataType):
                 '_fields_': fields,
             }
         )
-        return final_class, buffer
+
+        return final_class
 
     @classmethod
-    def to_python(cls, ctype_object, *args, **kwargs):
+    def to_python_not_null(cls, ctype_object, *args, **kwargs):
         result = []
         for i in range(ctype_object.length):
             result.append(
@@ -101,7 +103,7 @@ class ObjectArrayObject(GridGainDataType):
         return ctype_object.type_id, result
 
     @classmethod
-    def from_python(cls, value):
+    def from_python_not_null(cls, stream, value):
         type_or_id, value = value
         header_class = cls.build_header()
         header = header_class()
@@ -116,14 +118,13 @@ class ObjectArrayObject(GridGainDataType):
             length = 1
         header.length = length
         header.type_id = type_or_id
-        buffer = bytearray(header)
 
+        stream.write(header)
         for x in value:
-            buffer += infer_from_python(x)
-        return bytes(buffer)
+            infer_from_python(stream, x)
 
 
-class WrappedDataObject(GridGainDataType):
+class WrappedDataObject(GridGainDataType, Nullable):
     """
     One or more binary objects can be wrapped in an array. This allows reading,
     storing, passing and writing objects efficiently without understanding
@@ -149,10 +150,9 @@ class WrappedDataObject(GridGainDataType):
         )
 
     @classmethod
-    def parse(cls, client: 'Client'):
+    def parse_not_null(cls, stream):
         header_class = cls.build_header()
-        buffer = client.recv(ctypes.sizeof(header_class))
-        header = header_class.from_buffer_copy(buffer)
+        header = stream.read_ctype(header_class)
 
         final_class = type(
             cls.__name__,
@@ -165,21 +165,20 @@ class WrappedDataObject(GridGainDataType):
                 ],
             }
         )
-        buffer += client.recv(
-            ctypes.sizeof(final_class) - ctypes.sizeof(header_class)
-        )
-        return final_class, buffer
+
+        stream.seek(ctypes.sizeof(final_class), SEEK_CUR)
+        return final_class
 
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
         return bytes(ctype_object.payload), ctype_object.offset
 
     @classmethod
-    def from_python(cls, value):
+    def from_python(cls, stream, value):
         raise ParseError('Send unwrapped data.')
 
 
-class CollectionObject(GridGainDataType):
+class CollectionObject(GridGainDataType, Nullable):
     """
     Similar to object array, but contains platform-agnostic deserialization
     type hint instead of type ID.
@@ -242,15 +241,14 @@ class CollectionObject(GridGainDataType):
         )
 
     @classmethod
-    def parse(cls, client: 'Client'):
+    def parse_not_null(cls, stream):
         header_class = cls.build_header()
-        buffer = client.recv(ctypes.sizeof(header_class))
-        header = header_class.from_buffer_copy(buffer)
-        fields = []
+        header = stream.read_ctype(header_class)
+        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
 
+        fields = []
         for i in range(header.length):
-            c_type, buffer_fragment = AnyDataObject.parse(client)
-            buffer += buffer_fragment
+            c_type = AnyDataObject.parse(stream)
             fields.append(('element_{}'.format(i), c_type))
 
         final_class = type(
@@ -261,12 +259,15 @@ class CollectionObject(GridGainDataType):
                 '_fields_': fields,
             }
         )
-        return final_class, buffer
+        return final_class
 
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
         result = []
-        for i in range(ctype_object.length):
+        length = getattr(ctype_object, "length", None)
+        if length is None:
+            return None
+        for i in range(length):
             result.append(
                 AnyDataObject.to_python(
                     getattr(ctype_object, 'element_{}'.format(i)),
@@ -276,7 +277,7 @@ class CollectionObject(GridGainDataType):
         return ctype_object.type, result
 
     @classmethod
-    def from_python(cls, value):
+    def from_python_not_null(cls, stream, value):
         type_or_id, value = value
         header_class = cls.build_header()
         header = header_class()
@@ -291,14 +292,13 @@ class CollectionObject(GridGainDataType):
             length = 1
         header.length = length
         header.type = type_or_id
-        buffer = bytearray(header)
 
+        stream.write(header)
         for x in value:
-            buffer += infer_from_python(x)
-        return bytes(buffer)
+            infer_from_python(stream, x)
 
 
-class Map(GridGainDataType):
+class Map(GridGainDataType, Nullable):
     """
     Dictionary type, payload-only.
 
@@ -329,15 +329,14 @@ class Map(GridGainDataType):
         )
 
     @classmethod
-    def parse(cls, client: 'Client'):
+    def parse_not_null(cls, stream):
         header_class = cls.build_header()
-        buffer = client.recv(ctypes.sizeof(header_class))
-        header = header_class.from_buffer_copy(buffer)
-        fields = []
+        header = stream.read_ctype(header_class)
+        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
 
+        fields = []
         for i in range(header.length << 1):
-            c_type, buffer_fragment = AnyDataObject.parse(client)
-            buffer += buffer_fragment
+            c_type = AnyDataObject.parse(stream)
             fields.append(('element_{}'.format(i), c_type))
 
         final_class = type(
@@ -348,7 +347,7 @@ class Map(GridGainDataType):
                 '_fields_': fields,
             }
         )
-        return final_class, buffer
+        return final_class
 
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
@@ -368,7 +367,7 @@ class Map(GridGainDataType):
         return result
 
     @classmethod
-    def from_python(cls, value, type_id=None):
+    def from_python(cls, stream, value, type_id=None):
         header_class = cls.build_header()
         header = header_class()
         length = len(value)
@@ -380,12 +379,11 @@ class Map(GridGainDataType):
             )
         if hasattr(header, 'type'):
             header.type = type_id
-        buffer = bytearray(header)
 
+        stream.write(header)
         for k, v in value.items():
-            buffer += infer_from_python(k)
-            buffer += infer_from_python(v)
-        return bytes(buffer)
+            infer_from_python(stream, k)
+            infer_from_python(stream, v)
 
 
 class MapObject(Map):
@@ -420,17 +418,24 @@ class MapObject(Map):
 
     @classmethod
     def to_python(cls, ctype_object, *args, **kwargs):
-        return ctype_object.type, super().to_python(
+        obj_type = getattr(ctype_object, "type", None)
+        if obj_type is None:
+            return None
+        return obj_type, super().to_python(
             ctype_object, *args, **kwargs
         )
 
     @classmethod
-    def from_python(cls, value):
+    def from_python(cls, stream, value):
+        if value is None:
+            Null.from_python(stream)
+            return
+
         type_id, value = value
-        return super().from_python(value, type_id)
+        super().from_python(stream, value, type_id)
 
 
-class BinaryObject(GridGainDataType):
+class BinaryObject(GridGainDataType, Nullable):
     _type_id = TYPE_BINARY_OBJ
     type_code = TC_COMPLEX_OBJECT
 
@@ -442,42 +447,14 @@ class BinaryObject(GridGainDataType):
     COMPACT_FOOTER = 0x0020
 
     @staticmethod
-    def find_client():
-        """
-        A nice hack. Extracts the nearest `Client` instance from the
-        call stack.
-        """
-        from pygridgain import Client
-        from pygridgain.connection import Connection
-
-        frame = None
-        try:
-            for rec in inspect.stack()[2:]:
-                frame = rec[0]
-                code = frame.f_code
-                for varname in code.co_varnames:
-                    suspect = frame.f_locals[varname]
-                    if isinstance(suspect, Client):
-                        return suspect
-                    if isinstance(suspect, Connection):
-                        return suspect.client
-        finally:
-            del frame
-
-    @staticmethod
-    def hashcode(
-        value: object, client: 'Client' = None, *args, **kwargs
-    ) -> int:
+    def hashcode(value: object, client: None) -> int:
         # binary objects's hashcode implementation is special in the sense
         # that you need to fully serialize the object to calculate
         # its hashcode
-        if value._hashcode is None:
+        if not value._hashcode and client :
 
-            # …and for to serialize it you need a Client instance
-            if client is None:
-                client = BinaryObject.find_client()
-
-            value._build(client)
+            with BinaryStream(client.random_node) as stream:
+                value._from_python(stream, save_to_buf=True)
 
         return value._hashcode
 
@@ -525,37 +502,25 @@ class BinaryObject(GridGainDataType):
             },
         )
 
-    @staticmethod
-    def get_dataclass(conn: 'Connection', header) -> OrderedDict:
-        # get field names from outer space
-        result = conn.client.query_binary_type(
-            header.type_id,
-            header.schema_id
-        )
-        if not result:
-            raise ParseError('Binary type is not registered')
-        return result
-
     @classmethod
-    def parse(cls, client: 'Client'):
+    def parse_not_null(cls, stream):
         from pygridgain.datatypes import Struct
 
         header_class = cls.build_header()
-        buffer = client.recv(ctypes.sizeof(header_class))
-        header = header_class.from_buffer_copy(buffer)
+        header = stream.read_ctype(header_class)
+        stream.seek(ctypes.sizeof(header_class), SEEK_CUR)
 
         # ignore full schema, always retrieve fields' types and order
         # from complex types registry
-        data_class = cls.get_dataclass(client, header)
+        data_class = stream.get_dataclass(header)
         fields = data_class.schema.items()
         object_fields_struct = Struct(fields)
-        object_fields, object_fields_buffer = object_fields_struct.parse(client)
-        buffer += object_fields_buffer
+        object_fields = object_fields_struct.parse(stream)
         final_class_fields = [('object_fields', object_fields)]
 
         if header.flags & cls.HAS_SCHEMA:
             schema = cls.schema_type(header.flags) * len(fields)
-            buffer += client.recv(ctypes.sizeof(schema))
+            stream.seek(ctypes.sizeof(schema), SEEK_CUR)
             final_class_fields.append(('schema', schema))
 
         final_class = type(
@@ -567,19 +532,22 @@ class BinaryObject(GridGainDataType):
             }
         )
         # register schema encoding approach
-        client.compact_footer = bool(header.flags & cls.COMPACT_FOOTER)
-        return final_class, buffer
+        stream.compact_footer = bool(header.flags & cls.COMPACT_FOOTER)
+        return final_class
 
     @classmethod
     def to_python(cls, ctype_object, client: 'Client' = None, *args, **kwargs):
+        type_id = getattr(ctype_object, "type_id", None)
+        if type_id is None:
+            return None
 
         if not client:
             raise ParseError(
-                'Can not query binary type {}'.format(ctype_object.type_id)
+                'Can not query binary type {}'.format(type_id)
             )
 
         data_class = client.query_binary_type(
-            ctype_object.type_id,
+            type_id,
             ctype_object.schema_id
         )
         result = data_class()
@@ -595,21 +563,9 @@ class BinaryObject(GridGainDataType):
         return result
 
     @classmethod
-    def from_python(cls, value: object):
-
-        if getattr(value, '_buffer', None) is None:
-            client = cls.find_client()
-
-            # if no client can be found, the class of the `value` is discarded
-            # and the new dataclass is automatically registered later on
-            if client:
-                client.register_binary_type(value.__class__)
-            else:
-                raise Warning(
-                    'Can not register binary type {}'.format(value.type_name)
-                )
-
-            # build binary representation
-            value._build(client)
-
-        return value._buffer
+    def from_python_not_null(cls, stream, value):
+        if getattr(value, '_buffer', None):
+            stream.write(value._buffer)
+        else:
+            stream.register_binary_type(value.__class__)
+            value._from_python(stream)
